# This file is part of the graph-includes package
#
# (c) 2005 Yann Dirson <ydirson@altern.org>
# Distributed under version 2 of the GNU GPL.

package graphincludes::project;
use strict;
use warnings;

use graphincludes::params;

sub new {
  my $class = shift;
  my %args = @_;
  my $prefixstrip = $args{prefixstrip};
  my @files = @{$args{files}};	# take a copy of @ARGV
  my $self  = {};

  $self->{FILES}   = \@files;
  $self->{PFXSTRIP}= $prefixstrip;
  $self->{TREE}    = undef;
  $self->{NODEIDS} = undef;
  $self->{DEPS}    = undef;
  $self->{SPECIALEDGES} = {};
  $self->{REPORT} = { HDR => {},
		      SYS => {},
		    };
  $self->{_DROPCOUNT} = 0;

  bless ($self, $class);

  return $self;
}

sub init {
  my $self = shift;

  # store as a tree while keeping recording node id's
  $self->{NODEIDS} = $self->_nodeids($self->{TREE}, @{$self->{FILES}});
  $self->{DEPS} = $self->getdeps();
}

sub _nodeids {
  my $self = shift;
  my ($tree, @files) = @_;
  my $i=0;
  my %nodeid;
  foreach my $file (@files) {
    my $label = $self->filelabel($file);
    ## FIXME: why is return/assignment necessary !?
    #$tree = $self->insert ($label, $tree);
    if (!defined $nodeid{$label}) {
      $nodeid{$label} = "f" . ++$i;
      $nodeid{$label} .= '-' . $label if $graphincludes::params::debug;
    }
  }
  return \%nodeid;
}

sub locatefile {
  my $self = shift;
  my ($dst, @path) = @_;

  sub fullpath {
    my ($dstpath, $strip, $srcpath) = @_;
    join ('/', @$srcpath[0..($#$srcpath-$strip)], @$dstpath);
  }

  my @dstpath = split (m|/|, $dst);
  # count number of leading "../" in the #include reference
  my $strip = 0;
  while ($dstpath[0] eq '..') {
    $strip++; shift @dstpath;
  }
  # find the file, assuming lookup order should be -I. -I.. -I../.. etc...
  my $dstfile;
  foreach my $dir (@path) {
    my @srcpath = split (m|/|, $dir);
    if (defined($dstfile = fullpath(\@dstpath,$strip,\@srcpath)) and
	grep { $_ eq $dstfile } @{$self->{FILES}}) {
      last;
    } else {
      $dstfile = undef;
    }
  }

  return $dstfile;		# can be undef !
}

sub _suppress($$$@);

sub reduce {
  my $self = shift;
  my %reduceddeps = %{$self->{DEPS}};

  print STDERR "Doing transitive reduction " if $graphincludes::params::verbose;
  foreach my $node (keys %reduceddeps) {
    print STDERR '.' if $graphincludes::params::verbose;
    print STDERR "node $node\n" if $graphincludes::params::debug;
    if (defined $reduceddeps{$node}) {
      my @newdeps = @{$reduceddeps{$node}};
      my @considered = ($node);
      foreach my $child (@{$reduceddeps{$node}}) {
	# do not explore children already removed, or some circles cause lost edges
	next unless grep { $child eq $_ } @newdeps;
	print STDERR " child $child\n" if $graphincludes::params::debug;
	if ($child eq $node) {
	  # remove self-deps induced by trivial #include
	  print STDERR "  --$child\n" if $graphincludes::params::debug;
	  @newdeps = map { $_ eq $node ? () : $_ } @newdeps;
	} else {
	  if (defined $reduceddeps{$child}) {
	    foreach my $gchild (@{$reduceddeps{$child}}) {
	      if ($gchild ne $node and $gchild ne $child) { # XXX
		print STDERR "  gchild $gchild\n" if $graphincludes::params::debug;
		$self->_suppress (\@newdeps, \%reduceddeps, $gchild, \@considered, ($node, $child, $gchild));
	      }
	    }
	  }
	}
      }
      $reduceddeps{$node} = \@newdeps;
    }
  }
  print STDERR " $self->{_DROPCOUNT} cleared.\n" if $graphincludes::params::verbose;

  print STDERR "Verifying validity of transitive reduction " if $graphincludes::params::verbose;
  sub has_path {
    my ($graph, $from, $to, @seen) = @_;
    return 1 if $from eq $to;
    return 0 if grep { $_ eq $from } @seen;
    return 1 if grep { $_ eq $to } @{$graph->{$from}}; # superfluous ?
    foreach my $child (@{$graph->{$from}}) {
      return 1 if has_path($graph, $child, $to, (@seen, $from));
    }
    return 0;
  }
  foreach my $node (keys %{$self->{DEPS}}) {
    print STDERR '.' if $graphincludes::params::verbose;
    foreach my $child (@{$self->{DEPS}->{$node}}) {
      if (!has_path(\%reduceddeps, $node, $child)) {
	print STDERR "ERROR: missing edge from $node to $child\n";
      }
    }
  }
  print STDERR " done.\n" if $graphincludes::params::verbose;
  $self->{DEPS} = \%reduceddeps;
}

sub _suppress($$$@) {
  my $self = shift;
  my ($list,			# childlist to cleanup
      $reduceddeps,		# ref to hash to be reduced
      $node,			# node to consider this time
      $considered,		# graph nodes already seen, not to reconsider
      @context)			# current path, for printing only
    = @_;

  # Do not consider $node twice, prevent looping on circular deps.
  # We must take care of the special case of the child that led us to
  # the current node, or we would have to do special things to $gchild
  return if $node eq $context[1] or grep { $node eq $_ } (@$considered);
  push @$considered, $node;

  # remove $node from $list
  my @newlist = @$list;
  foreach my $child (@$list) {
    if ($child eq $node) {
      if ($graphincludes::params::showdropped) {
	$self->{SPECIALEDGES}->{$considered->[0]}->{$child} = 'color="#FFCCCC",constraint=false';
      } elsif (grep { $self->{NODEIDS}->{$_} eq $considered->[0] } @graphincludes::params::focus) {
	$self->{SPECIALEDGES}->{$considered->[0]}->{$child} = 'color="#FFCCCC"';
      } else {
	$self->{_DROPCOUNT}++;
	# FIXME: make more efficient ?
	@newlist = map { $_ eq $node ? () : $_ } @newlist;
      }
      print STDERR "    --$child (", join (',', @context), ")\n" if $graphincludes::params::debug;
      last;
    }
  }
  splice (@$list, 0, 0+@$list, @newlist);

  # remove $node's children
  if (defined $reduceddeps->{$node}) {
    foreach my $child (@{$reduceddeps->{$node}}) {
      if ($graphincludes::params::debug) {
	foreach (@context) {
	  print STDERR " ";
	}
	print STDERR "$child\n";
      }
      $self->_suppress ($list, $reduceddeps, $child, $considered, (@context, $child));
    }
  }
}

sub special_edge {
  my $self = shift;
  my ($src, $dst) = @_;

  my $attrs = $self->{SPECIALEDGES}->{$src}->{$dst};

  if (defined $attrs) {
    return "$attrs";
  } else {
    return undef;
  }
}

1;
